//
// gridcentric.cc
//
// NodeJS bindings for the libgridcentric guest library.
// This wraps all sensible function calls with callback 
// semantics, and keeps those with constant time as simple
// direct function calls.
// The event-driven callback semantics are actually fairly
// well matched to those function calls that require them.
//

#include <unistd.h>
#include <node/v8.h>
#include <node/node.h>

#include <gridcentric/gc-guest.h>

using namespace node;
using namespace v8;

#define REQUIRE(I, ISTYPE, CASTTYPE, NAME)                           \
  if( args.Length() <= (I) || !args[I]->Is##ISTYPE() )               \
    return ThrowException(Exception::TypeError(                      \
            String::New("Argument " #I " must be a " #ISTYPE "."))); \
  Local<CASTTYPE> NAME = Local<CASTTYPE>::Cast(args[I]);

#define REQUIRE_STRING(I, NAME) REQUIRE(I, String, String, NAME)
#define REQUIRE_INTEGER(I, NAME) REQUIRE(I, Number, Integer, NAME)
#define REQUIRE_CALLBACK(CLZ, I, NAME, DATA)           \
    REQUIRE(I, Function, Function, NAME);              \
    CallbackData##CLZ *DATA = new CallbackData##CLZ(); \
    DATA->This = args.This();                          \
    (void)DATA;

#define SET_VALUE(VAR, NAME, TYPE, VAL) VAR->Set(String::New(NAME), TYPE::New(VAL))
#define SET_INTEGER(VAR, NAME, VAL) SET_VALUE(VAR, NAME, Integer, VAL)
#define SET_STRING(VAR, NAME, VAL) SET_VALUE(VAR, NAME, String, VAL)
#define SET_DATE(VAR, NAME, VAL) SET_VALUE(VAR, NAME, Date, VAL)

class CallbackData {
public:
    Handle<Value> This;      // The this scope we were called in.
    Persistent<Function> cb; // The callback function passed.
    Handle<Value> rval;      // The return value to be given.
    CallbackData() : rval(Undefined()) { }
};

static int EIO_Post(eio_req *req)
{
    CallbackData *data = static_cast<CallbackData*>(req->data);
    ev_unref(EV_DEFAULT_UC);
    Local<Value> argv[1] = { *(data->rval) };
    TryCatch try_catch;
    data->cb->Call(Context::GetCurrent()->Global(), 1, argv);
    if (try_catch.HasCaught()) {
        FatalException(try_catch);
    }
    data->cb.Dispose();
    delete data;
    return 0;
}

#define DECL_CB_FIELDS(CLZ) \
class CallbackData##CLZ : public CallbackData { \
public:
#define END_CB_FIELDS \
};

#define DECL_CB_FUNC(CLZ) \
static int EIO_##CLZ(eio_req* req) { \
    CallbackData##CLZ *data = static_cast<CallbackData##CLZ*>(req->data); \
    (void)data;
#define END_CB_FUNC \
    return 0; \
}

#define DO_CALLBACK(CLZ, cb, data) \
    data->cb = Persistent<Function>::New(cb);               \
    eio_custom(EIO_##CLZ, EIO_PRI_DEFAULT, EIO_Post, data); \
    ev_ref(EV_DEFAULT_UC);                                  \
    return Undefined();

static Handle<Value> Version(const Arguments& args)
{
    HandleScope scope;
    Local<Integer> result = Integer::New(gc_version());
    return scope.Close(result);
}

static Handle<Value> Check(const Arguments& args)
{
    return Boolean::New((bool)gc_check());
}

static Handle<Value> VmId(const Arguments& args)
{
    HandleScope scope;
    Local<Integer> result = Integer::New(gc_vmid());
    return scope.Close(result);
}

static Handle<Value> IsMaster(const Arguments& args)
{
    return Boolean::New((bool)gc_is_master());
}

static Handle<Value> MyUuid(const Arguments& args)
{
    HandleScope scope;
    gc_uuid_t uuid;
    if( gc_my_uuid(&uuid) < 0 ) {
        return Undefined();
    } else {
        Local<String> result = String::New(uuid.value);
        return scope.Close(result);
    }
}

static Handle<Value> MyTicket(const Arguments& args)
{
    HandleScope scope;
    gc_uuid_t uuid;
    if( gc_my_ticket(&uuid) < 0 ) {
        Local<String> result = String::New(uuid.value);
        return scope.Close(result);
    } else {
        return Undefined();
    }
}

class DomainInfo {
public:
    static Persistent<FunctionTemplate> pt;

    static Handle<Object> Create(gc_domain_info_t info)
    {
        char strval[32];
        HandleScope scope;
        Local<Object> obj = Object::New();
        SET_STRING(obj, "id", info.id.value);
        SET_INTEGER(obj, "cpus", info.cpus);
        SET_INTEGER(obj, "ram", info.ram);
        SET_INTEGER(obj, "vmid", info.vmid);
        sprintf(strval, "%d.%d.%d.%d",
            (info.ip >> 24) & 0xff, (info.ip >> 16) & 0xff,
            (info.ip >> 8) & 0xff, info.ip & 0xff);
        SET_STRING(obj, "ip", strval);
        SET_STRING(obj, "ticket", info.ticket.value);
        return scope.Close(obj);
    }
};
Persistent<FunctionTemplate> DomainInfo::pt;

class DomainExitInfo {
public:
    static Persistent<FunctionTemplate> pt;

    static Handle<Object> Create(gc_domain_exit_t info)
    {
        HandleScope scope;
        Local<Object> obj = Object::New();
        SET_STRING(obj, "id", info.id.value);
        SET_INTEGER(obj, "exitcode", info.exitcode);
        return scope.Close(obj);
    }
};
Persistent<FunctionTemplate> DomainExitInfo::pt;

class TicketInfo {
public:
    static Persistent<FunctionTemplate> pt;

    static Handle<Object> Create(gc_ticket_info_t info)
    {
        HandleScope scope;
        Local<Object> obj = Object::New();
        SET_STRING(obj, "id", info.id.value);
        SET_STRING(obj, "status", gc_ticket_status_string(info.status));
        // We do not currently set the dates.
        // SET_DATE(obj, "start", info.start);
        // SET_DATE(obj, "end", info.end);
        SET_INTEGER(obj, "cpus", info.cpus);
        SET_INTEGER(obj, "vms", info.vms);
        SET_INTEGER(obj, "mincpuspervm", info.mincpuspervm);
        return scope.Close(obj);
    }
};
Persistent<FunctionTemplate> TicketInfo::pt;

template<class S, class D> Handle<Value> CreateArray(Handle<Value> This, S* src) {
    HandleScope scope;
    Local<Array> result = Array::New();
    S* cur = src;
    int i = 0;
    while( cur != NULL ) {
        result->Set(i, D::Create(*cur));
        cur = cur->next;
        i++;
    }
    return scope.Close(result);
}

DECL_CB_FIELDS(ListCloneDomains)
END_CB_FIELDS
DECL_CB_FUNC(ListCloneDomains)
    gc_domain_info_t* domains = gc_list_clone_domains();
    if( domains == GC_INVALID_DOMAIN_LIST ) {
        // Will be undefined by default.
    } else {
        // Construct a wrap a domain list.
        data->rval = CreateArray<gc_domain_info_t, DomainInfo>(data->This, domains);
    }
END_CB_FUNC

static Handle<Value> ListCloneDomains(const Arguments& args)
{
    REQUIRE_CALLBACK(ListCloneDomains, 0, cb, data);
    DO_CALLBACK(ListCloneDomains, cb, data);
}

DECL_CB_FIELDS(ListTicketDomains)
    gc_uuid_t ticket;
END_CB_FIELDS
DECL_CB_FUNC(ListTicketDomains)
    gc_domain_info_t* domains = gc_list_ticket_domains(data->ticket);
    if( domains == GC_INVALID_DOMAIN_LIST ) {
        // Will be undefined by default.
    } else {
        // Construct and wrap a domain list.
        data->rval = CreateArray<gc_domain_info_t, DomainInfo>(data->This, domains);
    }
END_CB_FUNC

static Handle<Value> ListTicketDomains(const Arguments& args)
{
    REQUIRE_STRING(0, ticket);
    REQUIRE_CALLBACK(ListTicketDomains, 1, cb, data);
    gc_init_uuid(&(data->ticket), *String::AsciiValue(ticket));
    DO_CALLBACK(ListTicketDomains, cb, data);
}

DECL_CB_FIELDS(GetDomainStatus)
    gc_uuid_t id;
END_CB_FIELDS
DECL_CB_FUNC(GetDomainStatus)
    gc_dom_status_t status = gc_get_domain_status(data->id);
    data->rval = String::New(gc_domain_status_string(status));
END_CB_FUNC

static Handle<Value> GetDomainStatus(const Arguments& args)
{
    REQUIRE_STRING(0, id);
    REQUIRE_CALLBACK(GetDomainStatus, 1, cb, data);
    gc_init_uuid(&(data->id), *String::AsciiValue(id));
    DO_CALLBACK(GetDomainStatus, cb, data);
}

DECL_CB_FIELDS(ListTickets)
END_CB_FIELDS
DECL_CB_FUNC(ListTickets)
    gc_ticket_info_t* tickets = gc_list_tickets();
    if( tickets == GC_INVALID_TICKET_LIST ) {
        // Already undefined.
    } else {
        // Convert to an array of TicketInfo objects.
        data->rval = CreateArray<gc_ticket_info_t, TicketInfo>(data->This, tickets);
    }
END_CB_FUNC

static Handle<Value> ListTickets(const Arguments& args)
{
    REQUIRE_CALLBACK(ListTickets, 0, cb, data);
    DO_CALLBACK(ListTickets, cb, data);
}

DECL_CB_FIELDS(WaitDomain)
    gc_uuid_t id;
END_CB_FIELDS
DECL_CB_FUNC(WaitDomain)
    gc_domain_exit_t status = gc_wait_domain(data->id);
    data->rval = DomainExitInfo::Create(status);
END_CB_FUNC

static Handle<Value> WaitDomain(const Arguments& args)
{
    REQUIRE_STRING(0, id);
    REQUIRE_CALLBACK(WaitDomain, 1, cb, data);
    gc_init_uuid(&(data->id), *String::AsciiValue(id));
    DO_CALLBACK(WaitDomain, cb, data);
}

DECL_CB_FIELDS(WaitAny)
END_CB_FIELDS
DECL_CB_FUNC(WaitAny)
    gc_domain_exit_t status = gc_wait_any();
    data->rval = DomainExitInfo::Create(status);
END_CB_FUNC

static Handle<Value> WaitAny(const Arguments& args)
{
    REQUIRE_CALLBACK(WaitAny, 0, cb, data);
    DO_CALLBACK(WaitAny, cb, data);
}

static Handle<Value> JoinParent(const Arguments& args)
{
    REQUIRE_INTEGER(0, code);
    // We do not expect to return from this function.
    gc_join_parent((int)(code->Value()));
    return Undefined();
}

static Handle<Value> JoinParentNoReport(const Arguments& args)
{
    // We do not expect to return from this function.
    gc_join_parent_noreport();
    return Undefined();
}

DECL_CB_FIELDS(SyncTicket)
    gc_uuid_t ticket;
    int timeout;
END_CB_FIELDS
DECL_CB_FUNC(SyncTicket)
    data->rval = Integer::New(gc_sync_ticket(data->ticket, data->timeout));
END_CB_FUNC

static Handle<Value> SyncTicket(const Arguments& args)
{
    REQUIRE_STRING(0, ticket);
    REQUIRE_INTEGER(1, timeout);
    REQUIRE_CALLBACK(SyncTicket, 2, cb, data);
    gc_init_uuid(&(data->ticket), *String::AsciiValue(ticket));
    data->timeout = timeout->Value();
    DO_CALLBACK(SyncTicket, cb, data);
}

DECL_CB_FIELDS(KillCloneDomain)
    gc_uuid_t id;
END_CB_FIELDS
DECL_CB_FUNC(KillCloneDomain)
    data->rval = Integer::New(gc_kill_clone_domain(data->id));
END_CB_FUNC

static Handle<Value> KillCloneDomain(const Arguments& args)
{
    REQUIRE_STRING(0, id);
    REQUIRE_CALLBACK(KillCloneDomain, 1, cb, data);
    gc_init_uuid(&(data->id), *String::AsciiValue(id));
    DO_CALLBACK(KillCloneDomain, cb, data);
}

DECL_CB_FIELDS(KillallCloneDomains)
END_CB_FIELDS
DECL_CB_FUNC(KillallCloneDomains)
    data->rval = Integer::New(gc_killall_clone_domains());
END_CB_FUNC

static Handle<Value> KillallCloneDomains(const Arguments& args)
{
    REQUIRE_CALLBACK(KillallCloneDomains, 0, cb, data);
    DO_CALLBACK(KillallCloneDomains, cb, data);
}

DECL_CB_FIELDS(KillallTicketDomains)
    gc_uuid_t ticket;
END_CB_FIELDS
DECL_CB_FUNC(KillallTicketDomains)
    data->rval = Integer::New(gc_killall_ticket_domains(data->ticket));
END_CB_FUNC

static Handle<Value> KillallTicketDomains(const Arguments& args)
{
    REQUIRE_STRING(0, ticket);
    REQUIRE_CALLBACK(KillallTicketDomains, 1, cb, data);
    gc_init_uuid(&(data->ticket), *String::AsciiValue(ticket));
    DO_CALLBACK(KillallTicketDomains, cb, data);
}

DECL_CB_FIELDS(RequestTicket)
    int maxcpus;
    int minvms;
    int mincpuspervm;
    int timeout;
END_CB_FIELDS
DECL_CB_FUNC(RequestTicket)
    gc_uuid_t uuid;
    if( gc_request_ticket(
            data->maxcpus, data->minvms, data->mincpuspervm,
            data->timeout, &uuid) < 0 ) {
        // Result is already undefined.
    } else {
        // Save the resulting ticket.
        data->rval = String::New(uuid.value);
    }
END_CB_FUNC

static Handle<Value> RequestTicket(const Arguments& args)
{
    REQUIRE_INTEGER(0, maxcpus);
    REQUIRE_INTEGER(1, minvms);
    REQUIRE_INTEGER(2, mincpuspervm);
    REQUIRE_INTEGER(3, timeout);
    REQUIRE_CALLBACK(RequestTicket, 4, cb, data);
    data->maxcpus = maxcpus->Value();
    data->minvms = minvms->Value();
    data->mincpuspervm = mincpuspervm->Value();
    data->timeout = timeout->Value();
    DO_CALLBACK(RequestTicket, cb, data);
}

DECL_CB_FIELDS(CancelTicket)
    gc_uuid_t ticket;
END_CB_FIELDS
DECL_CB_FUNC(CancelTicket)
    data->rval = Integer::New(gc_cancel_ticket(data->ticket));
END_CB_FUNC

static Handle<Value> CancelTicket(const Arguments& args)
{
    REQUIRE_STRING(0, ticket);
    REQUIRE_CALLBACK(CancelTicket, 1, cb, data);
    gc_init_uuid(&(data->ticket), *String::AsciiValue(ticket));
    DO_CALLBACK(CancelTicket, cb, data);
}

DECL_CB_FIELDS(Clone)
    gc_uuid_t ticket;
END_CB_FIELDS
DECL_CB_FUNC(Clone)
    data->rval = Integer::New(gc_clone(data->ticket));
END_CB_FUNC

static Handle<Value> Clone(const Arguments& args)
{
    REQUIRE_STRING(0, ticket);
    REQUIRE_CALLBACK(Clone, 1, cb, data);
    gc_init_uuid(&(data->ticket), *String::AsciiValue(ticket));
    DO_CALLBACK(Clone, cb, data);
}

DECL_CB_FIELDS(MakeTicketIndependent)
    gc_uuid_t ticket;
END_CB_FIELDS
DECL_CB_FUNC(MakeTicketIndependent)
    data->rval = Integer::New(gc_make_ticket_independent(data->ticket));
END_CB_FUNC

static Handle<Value> MakeTicketIndependent(const Arguments& args)
{
    REQUIRE_STRING(0, ticket);
    REQUIRE_CALLBACK(MakeTicketIndependent, 1, cb, data);
    gc_init_uuid(&(data->ticket), *String::AsciiValue(ticket));
    DO_CALLBACK(MakeTicketIndependent, cb, data);
}

static Handle<Value> MyIp(const Arguments& args)
{
    HandleScope scope;
    const char* res = gc_my_ip();
    if( res != NULL ) {
        Local<String> result = String::New(gc_my_ip());
        return scope.Close(result);
    } else {
        return Undefined();
    }
}

static Handle<Value> Reap(const Arguments& args)
{
    HandleScope scope;
    Local<Integer> result = Integer::New(gc_reap());
    return scope.Close(result);
}

#define BIND(x, y) \
    Local<FunctionTemplate> x = FunctionTemplate::New(y); \
    target->Set(String::NewSymbol(#x), x->GetFunction()); 

extern "C" {
  void init (Handle<Object> target)
  {
    BIND(version, Version);
    BIND(check, Check);
    BIND(vmid, VmId);
    BIND(is_master, IsMaster);
    BIND(request_ticket, RequestTicket);
    BIND(cancel_ticket, CancelTicket);
    BIND(clone, Clone);
    BIND(make_ticket_independent, MakeTicketIndependent);
    BIND(list_clone_domains, ListCloneDomains);
    BIND(list_ticket_domains, ListTicketDomains);
    BIND(get_domain_status, GetDomainStatus);
    BIND(list_tickets, ListTickets);
    BIND(wait_domain, WaitDomain);
    BIND(wait_any, WaitAny);
    BIND(sync_ticket, SyncTicket);
    BIND(join_parent, JoinParent);
    BIND(join_parent_noreport, JoinParentNoReport);
    BIND(kill_clone_domain, KillCloneDomain);
    BIND(killall_clone_domains, KillallCloneDomains);
    BIND(killall_ticket_domains, KillallTicketDomains);
    BIND(my_uuid, MyUuid);
    BIND(my_ticket, MyTicket);
    BIND(my_ip, MyIp);
    BIND(reap, Reap);
  }
}
